#include "stackpage.h"
#include "ui_stackpage.h"

#include <QDebug>
#include <QScrollBar>
#include <QGraphicsDropShadowEffect>
#include <QDir>
#include "baseuitr.h"

StackPage::StackPage(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::StackPage)
{
    ui->setupUi(this);

    QFile f(":/qss/baseui.qss");
    if(f.open(QIODevice::ReadOnly))
    {
        QString style = f.readAll().replace("\n", "");
        this->setStyleSheet(style);
        f.close();
    }

    _bottomLayout = new QHBoxLayout();
    _bottomLayout->setContentsMargins(0,0,0,0);
    _bottomLayout->setSpacing(20);
    ui->frame_buttom->setLayout( _bottomLayout );

    ui->btn_back->setText(g_tr_back);
    connect(ui->btn_back, &QToolButton::clicked, this, &StackPage::slotBack);
    connect(ui->scrollArea_central->verticalScrollBar(), &QScrollBar::valueChanged,
            this, &StackPage::slotScrollAreaValueChanged);

    //超时隐藏阴影
    _effectTimer = new QTimer(this);
    _effectTimer->setSingleShot(true);
    _effectTimer->setInterval(Qt::VeryCoarseTimer);
    _effectTimer->setInterval(3*1000);
    connect(_effectTimer, &QTimer::timeout, [=](){
        ui->frame_buttom->setGraphicsEffect(nullptr);
    });
}

StackPage::~StackPage()
{
    delete ui;
}

void StackPage::refreshStackPage()
{
    this->getTopPage()->refreshPage();
}

Page *StackPage::getPage(const QString key)
{
    QMap<QString, Page*>::iterator it = _pages.find(key);
    if(it != _pages.end())
    {
        return it.value();
    }
    return nullptr;
}

Page *StackPage::getTopPage()
{
    Q_ASSERT(_stackPage.count() > 0);
    return _stackPage.top();
}

void StackPage::setFirstPage(const QString key)
{
    this->slotNextPage(key);
}

void StackPage::setNextPage(const QString key)
{
    this->slotNextPage(key);
}

void StackPage::setSwitchPage(const QString key)
{
    this->slotSwitchPage(key);
}

void StackPage::slotBackPage()
{
    Q_ASSERT(_stackPage.count() > 1);

    //back 之前应该做的一切操作
    Page * page = _stackPage.top();
    page->backBtnClicked();

    this->removeTopPage();
    this->updateTopPage();
}

void StackPage::slotNextPage(const QString key)
{
    QMap<QString, Page*>::iterator it = _pages.find(key);
    if(it != _pages.end())
    {
        Page * page = it.value();
        this->addTopPage(page);
        this->updateTopPage();
    }
    else {
        Page * page = createPage(key);

        Q_ASSERT(page != nullptr);

        _pages.insert(key, page);
        this->addTopPage(page);
        this->updateTopPage();
    }
}

void StackPage::slotSwitchPage(const QString key)
{
    this->removeTopPage();

    QMap<QString, Page*>::iterator it = _pages.find(key);
    if(it != _pages.end())
    {
        Page * page = it.value();
        this->addTopPage(page);
        this->updateTopPage();
    }
    else
    {
        Page * page = createPage(key);

        Q_ASSERT(page != nullptr);

        _pages.insert(key, page);
        this->addTopPage(page);
        this->updateTopPage();
    }
}

void StackPage::slotPageTitleChanged(QString text)
{
    ui->label_title->setText(text);
}

void StackPage::slotButtomBtnVisibleChanged(int key, bool visible)
{
    bool widget_buttom_visible = false;
    //
    for(QToolButton * btn : _bottomBtns)
    {
        int btnKey = btn->property("PAGE_BBTN_KEY").toInt();
        if(btnKey == key)
        {
            if(visible)
            {
                //需要先设置父widget的显示，子widget的显示才能生效
                ui->frame_buttom->setVisible( true );
                btn->setVisible(true);
            }
            else
            {
                btn->setVisible(false);
            }
        }

        if(btn->isVisible()) widget_buttom_visible = true;
    }
    ui->frame_buttom->setVisible( widget_buttom_visible );
}

void StackPage::slotBack()
{
    ui->btn_back->disconnect();

    Q_ASSERT(_stackPage.count() > 1);

    //back 之前应该做的一切操作
    Page * page = _stackPage.top();
    page->backBtnClicked();

    this->removeTopPage();
    this->updateTopPage();

    connect(ui->btn_back, &QToolButton::clicked, this, &StackPage::slotBack);
}

void StackPage::slotBottom()
{
    if(_stackPage.count() <= 0) return;

    QToolButton * btn = qobject_cast<QToolButton *>(sender());
    if(btn)
    {
        btn->disconnect();

        int key = btn->property("PAGE_BBTN_KEY").toInt();
        Page * page = _stackPage.top();
        if(key == Page::PAGE_BBTN_APPLY)
        {
            page->applyBtnClicked();
        }
        else
        {
            page->buttomBtnClicked(key);
        }
        connect(btn, &QToolButton::clicked, this, &StackPage::slotBottom);
    }
}

void StackPage::slotScrollAreaValueChanged(int value)
{
    Q_UNUSED(value)

    if(ui->frame_buttom->isVisible())
    {
        if(_effectTimer->isActive())
        {
            _effectTimer->stop();
            _effectTimer->start();
        }
        else
        {
            _effectTimer->start();
        }

        //设置阴影
        if(!ui->frame_buttom->graphicsEffect())
        {
            QGraphicsDropShadowEffect * effect = new QGraphicsDropShadowEffect();
            effect->setOffset(0,-3);
            QColor color1("#000000");
            color1.setAlphaF(0.2);
            effect->setColor(color1);
            effect->setBlurRadius(20);

            ui->frame_buttom->setGraphicsEffect(effect);
        }
    }

//    int step = 3;
//    int v = (TITLE_MAX_HEIGHT - TITLE_MIN_HEIGHT) * step;
//    if(value <= v)
//    {
//        ui->widget_top->setFixedHeight(TITLE_MAX_HEIGHT - value/step);
//    }
//    else
//    {
//        ui->widget_top->setFixedHeight(TITLE_MIN_HEIGHT);
//    }
}

void StackPage::addTopPage(Page *page)
{
    Q_ASSERT(page);

    //disconnect
    if(_stackPage.count() > 0)
    {
        Page* lastPage = _stackPage.top();
        lastPage->hide();

        disconnect(lastPage, &Page::backPage, this, &StackPage::slotBackPage);
        disconnect(lastPage, &Page::pageTitleChanged, this, &StackPage::slotPageTitleChanged);
        disconnect(lastPage, &Page::nextPage, this, &StackPage::slotNextPage);
        disconnect(lastPage, &Page::switchPage, this, &StackPage::slotSwitchPage);
        disconnect(lastPage, &Page::buttomBtnVisibleChanged, this, &StackPage::slotButtomBtnVisibleChanged);
    }

    _stackPage.push(page);
}

void StackPage::removeTopPage()
{
    Q_ASSERT(_stackPage.count() > 1);

    Page * lastPage = _stackPage.pop();

    lastPage->hide();

    disconnect(lastPage, &Page::backPage, this, &StackPage::slotBackPage);
    disconnect(lastPage, &Page::pageTitleChanged, this, &StackPage::slotPageTitleChanged);
    disconnect(lastPage, &Page::nextPage, this, &StackPage::slotNextPage);
    disconnect(lastPage, &Page::switchPage, this, &StackPage::slotSwitchPage);
    disconnect(lastPage, &Page::buttomBtnVisibleChanged, this, &StackPage::slotButtomBtnVisibleChanged);
}

void StackPage::updateTopPage()
{
    Q_ASSERT(_stackPage.count() > 0);

    Page * page = _stackPage.top();
    ui->scrollArea_central->takeWidget();
    ui->scrollArea_central->setWidget(page);

    //connect
    connect(page, &Page::backPage, this, &StackPage::slotBackPage);
    connect(page, &Page::pageTitleChanged, this, &StackPage::slotPageTitleChanged);
    connect(page, &Page::nextPage, this, &StackPage::slotNextPage);
    connect(page, &Page::switchPage, this, &StackPage::slotSwitchPage);
    connect(page, &Page::buttomBtnVisibleChanged, this, &StackPage::slotButtomBtnVisibleChanged);

    this->setPageTitle();
    this->setBackBtnVisible();

    this->clearBottomBtns();
    this->addBottomBtns();
    this->setBottomBtnsVisible();

    page->show();
    page->refreshPage();
}

void StackPage::setPageTitle()
{
    Q_ASSERT(_stackPage.count() > 0);
    Page *page = _stackPage.top();
    ui->label_title->setText(page->pageTitle());
}

void StackPage::setBackBtnVisible()
{
    bool v = (_stackPage.count() > 1 ? true:false);
    ui->btn_back->setVisible(v);
}

void StackPage::addBottomBtns()
{
    Q_ASSERT(_stackPage.count() > 0);
    Page *page = _stackPage.top();
    //QPair<key, text>
    QList<QPair<int, QString> > pairs = page->buttomBtnsPair();
    QList<int> redButtonKeys = page->buttomBtnsRed();

    if(pairs.count()) _bottomLayout->addStretch();
    for(QPair<int, QString> pair : pairs)
    {
        QToolButton * btn = new QToolButton(ui->frame_buttom);
        btn->setProperty("ShowRed", redButtonKeys.contains(pair.first));
        btn->setFixedHeight(30);
        btn->setMinimumWidth(120);
        btn->setProperty("PAGE_BBTN_KEY", pair.first);
        btn->setText(pair.second);
        connect(btn, &QToolButton::clicked, this, &StackPage::slotBottom);

        _bottomBtns.append(btn);
        _bottomLayout->addWidget(btn);
    }
    if(pairs.count()) _bottomLayout->addStretch();
}

void StackPage::setBottomBtnsVisible()
{
    Q_ASSERT(_stackPage.count() > 0);
    Page *page = _stackPage.top();
    QList<int> vs = page->buttomBtnsVisible();

    if(vs.count() && _bottomBtns.count())
    {
        //需要先设置父widget的显示，子widget的setvisible才能生效
        ui->frame_buttom->setVisible( true );
        for(QToolButton *btn : _bottomBtns)
        {
            int key = btn->property("PAGE_BBTN_KEY").toInt();
            btn->setVisible(vs.contains(key));
        }
    }
    else
    {
        ui->frame_buttom->setVisible( false );
    }
}

void StackPage::clearBottomBtns()
{
    while (_bottomBtns.count()) {
        QToolButton * btn = _bottomBtns.takeLast();
        btn->disconnect();
        _bottomLayout->removeWidget(btn);
        _bottomBtns.removeAll(btn);
        btn->deleteLater();
    }
    while (_bottomLayout->count()) {
        QLayoutItem * item = _bottomLayout->takeAt(0);
        if(item->widget()) item->widget()->deleteLater();
        delete item;
    }

    //
    ui->frame_buttom->setVisible( false );
}
