#include "monitoroutputwidget.h"
#include <KScreen/EDID>
#include <KScreen/Output>
#include <KScreen/ConfigMonitor>
#include <QFile>
#include <Nde/Settings>
#include <KScreen/Config>
#include <QByteArray>
#include <QProcess>
#include <iostream>
#include "monitor.h"
#include <random>

QString modeToString(KScreen::ModePtr mode)
{
    return QString("%1x%2").arg(mode->size().width()).arg(mode->size().height());
}

KScreen::OutputPtr getOutputById(int id, KScreen::OutputList outputs)
{
    for (const KScreen::OutputPtr &output : outputs)
        if (output->id() == id)
            return output;

    return KScreen::OutputPtr(nullptr);
}

KScreen::ModePtr getModeById(QString id, KScreen::ModeList modes)
{
    for (const KScreen::ModePtr &mode : modes)
        if (mode->id() == id)
            return mode;

    return KScreen::ModePtr(nullptr);
}
MonitorOutputWidget::MonitorOutputWidget(KScreen::ConfigPtr config, QWidget *parent) :

    QWidget(parent),
    mConfig(config),
    ui(new Ui::MonitorOutputWidget),
    mBrightness(new XRandrBrightness () )
{
    ui->setupUi(this);

    QFile f(":/qss/monitoroutputwidget.qss");
    if(f.open(QIODevice::ReadOnly))
    {
        QString style = f.readAll().replace("\n", "");
        this->setStyleSheet(style);
    }

    connect(KScreen::ConfigMonitor::instance(), &KScreen::ConfigMonitor::configurationChanged, this, &MonitorOutputWidget::initOutput);
    initOutput();
}
void MonitorOutputWidget::initOutput()
{
    int currentModeId = ui->displayCombo->currentData().toInt();
    bool exist = false;
    disconnect(ui->displayCombo, SIGNAL(currentIndexChanged(int)), this, nullptr);
    ui->displayCombo->clear();

    mOutputs = mConfig->outputs();
    for (auto i = mOutputs.begin(); i != mOutputs.end(); ++i) {
        int key = i.key();
        KScreen::OutputPtr output = i.value();
        if(!output->isConnected())
            continue;
        if(currentModeId == key)
            exist = true;
        QString monitorName = output->edid()->name();
        if(monitorName.trimmed().size() > 0)
            monitorName = output->name() + " (" + monitorName + ")";
        else {
            monitorName = output->name();
        }
        ui->displayCombo->addItem(monitorName, key);
    }
    if(ui->displayCombo->count() <= 0)
	return;
    connect(ui->displayCombo, QOverload<int>::of(&QComboBox::currentIndexChanged), this, &MonitorOutputWidget::outputChanged);
    int index = 0;
    if(exist) {
        index = ui->displayCombo->findData(currentModeId);
    }
    ui->displayCombo->setCurrentIndex(index);
    outputChanged(index);
}
void MonitorOutputWidget::outputChanged(int index)
{
    if(index < 0)
        return;

    KScreen::OutputPtr output = mOutputs[ui->displayCombo->itemData(index).toInt()];
    updateDisplay(output);
}
void MonitorOutputWidget::updateDisplay(const KScreen::OutputPtr &output)
{
    disconnect(ui->primaryCheckbox, SIGNAL(toggled(bool)), this, nullptr);
    disconnect(ui->enabledCheckbox, SIGNAL(toggled(bool)), this, nullptr);
    ui->primaryCheckbox->setChecked(output->isPrimary());
    ui->enabledCheckbox->setChecked(output->isEnabled());
    connect(ui->primaryCheckbox, SIGNAL(toggled(bool)), this, SLOT(onPrimaryChanged(bool)));
    connect(ui->enabledCheckbox, SIGNAL(toggled(bool)), this, SLOT(onEnabledChanged(bool)));

    //多屏
    initMutilScreen();
    //分辨率
    QList <KScreen::ModePtr> modeList = output->modes().values();
    // Remove duplicate sizes
    QMap<QString, KScreen::ModePtr> noDuplicateModes;
    for(const KScreen::ModePtr &mode : qAsConst(modeList))
    {
        if(mCurrentMutilMode == UNIFIED_SCREEN) {
            QList<QSize> commonSizes = getCommonSizes();
            if(commonSizes.contains(mode->size())){
                noDuplicateModes[modeToString(mode)] = mode;
            }
        }else {
            if( noDuplicateModes.keys().contains(modeToString(mode)) )
            {
                if(output->currentModeId() == mode->id())
                    noDuplicateModes[modeToString(mode)] = mode;
            }
            else
                noDuplicateModes[modeToString(mode)] = mode;
        }
    }

    modeList = noDuplicateModes.values();
    qSort(modeList.begin(), modeList.end(), sizeBiggerThan);

    const char* home = getenv("HOME");
#define DEFAULT_CONFIG ".config/nde/zte-monitor.conf"
    if (QFile::exists(QString::fromLocal8Bit(home) + QLatin1Char('/') +QLatin1String(DEFAULT_CONFIG)))
    {
        Nde::Settings settings("zte-monitor");
        bool EnableresolutionCombo = settings.value("EnableresolutionCombo").toBool();
        if(!EnableresolutionCombo)
        {
            ui->resolutionCombo->setEnabled(false);
        }
    }
    disconnect(ui->resolutionCombo, SIGNAL(currentIndexChanged(int)), this, nullptr);
    ui->resolutionCombo->clear();
    // Add each mode to the list
    for (const KScreen::ModePtr &mode : qAsConst(modeList))
    {
        ui->resolutionCombo->addItem(modeToString(mode), mode->id());
    }
    // Select actual mode in list
    if (output->currentMode())
    {
        // Set the current mode in dropdown
        int idx = ui->resolutionCombo->findData(output->currentMode()->id());
        if (idx < 0)
        {
            // Select mode with same size
            for (const KScreen::ModePtr &mode : output->modes().values())
            {
                if( mode->size() == output->currentMode()->size() )
                    idx = ui->resolutionCombo->findData(output->currentMode()->id());
            }
        }
        if(idx < 0)
            idx = ui->resolutionCombo->findData(output->preferredMode()->id());
        if(idx >= 0) {
            ui->resolutionCombo->setCurrentIndex(idx);
            QFont font = ui->resolutionCombo->font();
            font.setBold(true);
            ui->resolutionCombo->setItemData(idx, font, Qt::FontRole);
        }
    }
    connect(ui->resolutionCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(onResolutionChanged(int)));
    updateRefreshRates();

    //方向
    disconnect(ui->orientationCombo, SIGNAL(currentIndexChanged(int)), this, nullptr);
    ui->orientationCombo->clear();
    // Insert orientations
    ui->orientationCombo->addItem(tr("None"), KScreen::Output::None);
    ui->orientationCombo->addItem(tr("Left"), KScreen::Output::Left);
    ui->orientationCombo->addItem(tr("Right"), KScreen::Output::Right);
    ui->orientationCombo->addItem(tr("Inverted"), KScreen::Output::Inverted);
    switch(output->rotation())
    {
    case KScreen::Output::None:
        ui->orientationCombo->setCurrentIndex(0);
        break;
    case KScreen::Output::Left:
        ui->orientationCombo->setCurrentIndex(1);
        break;
    case KScreen::Output::Right:
        ui->orientationCombo->setCurrentIndex(2);
        break;
    case KScreen::Output::Inverted:
        ui->orientationCombo->setCurrentIndex(3);
        break;
    }
    connect(ui->orientationCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(onOrientationChanged(int)));
    //屏幕亮度
    if(mBrightness->canBrightness())
    {
        mBrightness->setOutput(output);
        ui->brightnessSlider->setValue(brighttovalue(mBrightness->gamma_brightness_get() *  100));
        connect(ui->brightnessSlider, &QSlider::valueChanged, this, [&](int value){
            mBrightness->gamma_brightness_set(valuetobright(value)/100);
        });
    }
    //缩放屏幕
    disconnect(ui->dpiScalePrecent, SIGNAL(currentIndexChanged(int)), this, nullptr);
    ui->dpiScalePrecent->clear();
    ui->dpiScalePrecent->addItem(tr("100%"), 96);
    ui->dpiScalePrecent->addItem(tr("200%"), 192);
    if(ui->dpiScalePrecent->itemData(0).toInt() == getCurrentDpi())
        ui->dpiScalePrecent->setCurrentIndex(0);
    else
        ui->dpiScalePrecent->setCurrentIndex(1);
    connect(ui->dpiScalePrecent, SIGNAL(currentIndexChanged(int)), this, SLOT(onDpiScalePrecentChanged(int)));
}

int MonitorOutputWidget::brighttovalue(float f)
{
    return static_cast<int>((f-30)/0.7);
}

float MonitorOutputWidget::valuetobright(int v)
{
    return 30 + 0.7 * v;
}
void MonitorOutputWidget::initMutilScreen()
{
    disconnect(ui->mutilScreenCombo, SIGNAL(currentIndexChanged(int)), this, nullptr);
    ui->mutilScreenCombo->clear();

    int screenCount = mConfig->connectedOutputs().count();
    if(screenCount > 1)
    {
        ui->mutilScreenCombo->addItem(tr("Copy these displays"));
        ui->mutilScreenCombo->addItem(tr("Extend these displays"));
        //暂定双屏
        for(int i = 0 ; i < screenCount; i++)
        {
            KScreen::OutputPtr output = mOutputs[ui->displayCombo->itemData(i).toInt()];
            ui->mutilScreenCombo->addItem(tr("Show only on %1").arg(output->name()), output->name());
        }
        bool isEnabled = true;
        KScreen::OutputPtr enabledOutput;
        for (KScreen::OutputPtr output : mOutputs) {
            if(output->isConnected() && !output->isEnabled())
                isEnabled = false;
            else if(output->isConnected())
                enabledOutput = output;
        }
        if(!isEnabled){
            int index = ui->mutilScreenCombo->findData(enabledOutput->name());
            if(index > 0) {
                ui->mutilScreenCombo->setCurrentIndex(index);
                mCurrentMutilMode = ONLY_1_SCREEN;
            }
        } else {
            for (int i = 0 ; i < screenCount; i++) {
                KScreen::OutputPtr output1 = mOutputs[ui->displayCombo->itemData(i).toInt()];
                if(enabledOutput->name() == output1->name())
                    continue;
                if(enabledOutput->pos() == output1->pos()) {
                    ui->mutilScreenCombo->setCurrentIndex(0);
                    mCurrentMutilMode = UNIFIED_SCREEN;
                }
                else {
                    ui->mutilScreenCombo->setCurrentIndex(1);
                    mCurrentMutilMode = EXTENDED_SCREEN;
                }
            }
        }

        connect(ui->mutilScreenCombo, QOverload<int>::of(&QComboBox::currentIndexChanged), this, &MonitorOutputWidget::onMutilScreenIndexChanged);
        ui->mutilScreenFrame->setVisible(true);
        ui->enabled_display->setVisible(true);
        ui->enabled_primary->setVisible(true);
    }
    else {
        ui->mutilScreenFrame->setVisible(false);
        ui->enabled_display->setVisible(false);
        ui->enabled_primary->setVisible(false);
    }
}
int MonitorOutputWidget::getCurrentDpi()
{
    QProcess process;
    process.start("xrdb -query");
    process.waitForFinished();
    while(!process.atEnd()) {
        QString dpi = QString(process.readLine());
        QStringList dpiList =  dpi.split(QStringLiteral(":"));
        if(dpiList.size() == 2 &&  dpiList.at(0).trimmed() == "Xft.dpi") {
            return dpiList.at(1).trimmed().toInt();
        }
    }
    return 0;
}
void MonitorOutputWidget::onResolutionChanged(int index)
{
    if(mCurrentMutilMode == UNIFIED_SCREEN) {
        for (const KScreen::OutputPtr &output : mOutputs) {
            QList <KScreen::ModePtr> modeList = output->modes().values();
            QString currentResolution = ui->resolutionCombo->currentText();
            for(const KScreen::ModePtr &mode : qAsConst(modeList)) {
                if(modeToString(mode) == currentResolution)
                    output->setCurrentModeId(mode->id());
            }
        }
    }else if(mCurrentMutilMode == EXTENDED_SCREEN) {
        KScreen::OutputPtr currentOutput = mOutputs[ui->displayCombo->currentData().toInt()];
        currentOutput->setCurrentModeId(ui->resolutionCombo->currentData().toString());
        for (const KScreen::OutputPtr &output : mOutputs) {
            if(output->id() != currentOutput->id() &&
                    output->pos().x() > currentOutput->pos().x()) {
                output->setPos(QPoint(currentOutput->size().width(), 0));
            }
        }
    }
    else {
        KScreen::OutputPtr output = mOutputs[ui->displayCombo->currentData().toInt()];
        output->setCurrentModeId(ui->resolutionCombo->currentData().toString());
    }

    updateRefreshRates();
}
void MonitorOutputWidget::positionChanged()
{
    mCurrentMutilMode = OTHER_SCREEN;
    int currentModeId = ui->displayCombo->currentData().toInt();
    int index = ui->displayCombo->findData(currentModeId);
    outputChanged(index);
}
void MonitorOutputWidget::onMutilScreenIndexChanged(int index)
{
    switch (index) {
    case UNIFIED_SCREEN:
        mCurrentMutilMode = UNIFIED_SCREEN;
        unified();
        break;
    case EXTENDED_SCREEN:
        mCurrentMutilMode = EXTENDED_SCREEN;
        extended();
        positionChanged();
        break;
    default:
        mCurrentMutilMode = ONLY_1_SCREEN;
        onlyOne();
        positionChanged();
        break;
    }

}
void MonitorOutputWidget::extended()
{
    int width = 0;
    KScreen::OutputList outputs = mConfig->outputs();
    for (const KScreen::OutputPtr &output : outputs)
    {
        if( !output->isConnected() )
            continue;
        QPoint pos = output->pos();
        pos.setX(width);
        pos.setY(0);
        output->setPos(pos);
        output->setEnabled(true);
        KScreen::ModePtr mode = output->preferredMode();
        output->setCurrentModeId(mode->id());
        if (mode) {
            width += mode->size().width();
        }
    }
}

static bool sizeThan(const QSize &sizeA, const QSize &sizeB)
{
    return sizeA.width() * sizeA.height() > sizeB.width() * sizeB.height();
}
QList<QSize> MonitorOutputWidget::getCommonSizes()
{
    const KScreen::OutputList outputs = mConfig->outputs();
    // Look for common size
    QList<QSize> commonSizes;
    for (const KScreen::OutputPtr &output : outputs)
    {
        if( !output->isConnected() )
            continue;

        const auto modes = output->modes();
        for(const KScreen::ModePtr &mode : modes)
        {
            commonSizes.append(mode->size());
        }
        break;
    }
    for (const KScreen::OutputPtr &output : outputs)
    {
        if( !output->isConnected() )
            continue;
        QList<QSize> sizes;
        const auto modes = output->modes();
        for(const KScreen::ModePtr &mode : modes)
        {
            if( commonSizes.contains(mode->size()) )
                sizes.append(mode->size());
        }
        commonSizes = sizes;
    }
    // Select the bigest common size
    qSort(commonSizes.begin(), commonSizes.end(), sizeThan);
    return commonSizes;
}
void MonitorOutputWidget::unified()
{
    const KScreen::OutputList outputs = mConfig->outputs();
    QList<QSize> commonSizes  = getCommonSizes();
    if(commonSizes.isEmpty())
        return;
    QSize commonSize = commonSizes[0];
    // Put all monitors in (0,0) position and set size
    for (const KScreen::OutputPtr &output : outputs)
    {
        if( !output->isConnected() )
            continue;
        QPoint pos = output->pos();
        pos.setX(0);
        pos.setY(0);
        output->setPos(pos);
        output->setEnabled(true);
        // Select mode with the biggest refresh rate
        float maxRefreshRate = 0.0;
        const auto outputs = output->modes();
        for(const KScreen::ModePtr &mode : outputs)
        {
            if(mode->size() == commonSize && maxRefreshRate < mode->refreshRate())
            {
                output->setCurrentModeId(mode->id());
                maxRefreshRate = mode->refreshRate();
            }
        }
    }
}
void MonitorOutputWidget::onlyOne()
{
    QString name = ui->mutilScreenCombo->currentData().toString();
    const KScreen::OutputList outputs = mConfig->outputs();
    for (const KScreen::OutputPtr &output : outputs)
    {
        if(!output->isConnected() )
            continue;
        if(output->name() == name) {
            if(!output->isEnabled())
                output->setCurrentModeId(output->preferredModeId());
            output->setEnabled(true);
        }
        else
            output->setEnabled(false);
    }
}
void MonitorOutputWidget::onPrimaryChanged(bool enabled)
{
    KScreen::OutputPtr output = mOutputs[ui->displayCombo->currentData().toInt()];
    output->setPrimary(enabled);
}
void MonitorOutputWidget::onEnabledChanged(bool enabled)
{
    KScreen::OutputPtr output = mOutputs[ui->displayCombo->currentData().toInt()];
    int enabledCount = 0;
    for (const KScreen::OutputPtr &o : mOutputs) {
        if( !o->isConnected() )
            continue;
        if(o == output)
            continue;
        if(o->isEnabled()) {
            enabledCount ++;
        }
    }
    if(enabledCount == 0) {
        ui->enabledCheckbox->setChecked(true);
        return;
    }
    output->setEnabled(enabled);

    // If we're enabling a disabled output for the first time
    if (enabled && !output->currentMode())
    {

        onResolutionChanged(ui->resolutionCombo->currentIndex());
        onOrientationChanged(ui->orientationCombo->currentIndex());
        int width = 0;
        for (const KScreen::OutputPtr &o : mOutputs) {
            if( !o->isConnected() )
                continue;
            if(o == output)
                continue;
            width += o->currentMode()->size().width();
        }
        QPoint pos = output->pos();
        pos.setX(width);
        pos.setY(0);
        output->setPos(pos);
    }
    positionChanged();
}
void MonitorOutputWidget::onDpiScalePrecentChanged(int index)
{
   static QMap<int, int> scaleMap = {{1,96}, {2, 192}, {3, 288}};
   const QByteArray input = "Xft.dpi: " + QByteArray::number(scaleMap[index + 1]);
   QProcess p;
   p.start(QStringLiteral("xrdb"), { QStringLiteral("-quiet"), QStringLiteral("-merge"), QStringLiteral("-nocpp") });
   p.setProcessChannelMode(QProcess::ForwardedChannels);
   p.write(input);
   p.closeWriteChannel();
   p.waitForFinished(-1);
}
void MonitorOutputWidget::onOrientationChanged(int index)
{
    KScreen::OutputPtr output = mOutputs[ui->displayCombo->currentData().toInt()];
    output->setRotation(static_cast<KScreen::Output::Rotation>(ui->orientationCombo->currentData().toInt()));
}
void MonitorOutputWidget::updateRefreshRates()
{
    KScreen::OutputPtr output = mOutputs[ui->displayCombo->currentData().toInt()];
    disconnect(ui->rateCombo, SIGNAL(currentIndexChanged(int)), this, nullptr);
    ui->rateCombo->clear();
    if (output->modes().size() < 0)
        return;
    KScreen::ModePtr selectedMode = output->currentMode();
    if (selectedMode)
    {
        const auto modes = output->modes();
        for (const KScreen::ModePtr &mode : modes) {
            if (mode->size() == selectedMode->size()){
                ui->rateCombo->addItem(tr("%1 Hz").arg(mode->refreshRate()), mode->id());
            }
        }
        int idx = ui->rateCombo->findData(selectedMode->id());
        if (idx >= 0)
            ui->rateCombo->setCurrentIndex(idx);
    }

    connect(ui->rateCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(onRateChanged(int)));
}
void MonitorOutputWidget::onRateChanged(int index)
{
    KScreen::OutputPtr output = mOutputs[ui->displayCombo->currentData().toInt()];
    output->setCurrentModeId(ui->rateCombo->currentData().toString());
}

MonitorOutputWidget::~MonitorOutputWidget()
{
    delete ui;
    delete mBrightness;
}
