/*
 *    Copyright (C) 2014  P.L. Lucas <selairi@gmail.com>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License along
 *    with this program; if not, write to the Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "monitorpicture.h"

#include <QFont>
#include <QFontMetrics>
#include <QPen>
#include <QDebug>
#include <QVector2D>
#include <QScrollBar>
#include <KScreen/Types>
#include <KScreen/Screen>
#include <KScreen/Config>
#include <KScreen/ConfigMonitor>
#include <iostream>
#include <QGraphicsSceneMouseEvent>
bool comparePositon(MonitorPicture *monitorPicture1, MonitorPicture *monitorPicture2)
{
    return monitorPicture1->originX < monitorPicture2->originX ||
            monitorPicture1->originY < monitorPicture2->originY;
}
MonitorPictureProxy::MonitorPictureProxy(QObject *parent, MonitorPicture *monitorPicture):QObject(parent)
{
    this->monitorPicture = monitorPicture;
}

void MonitorPictureProxy::updateRotation()
{
    monitorPicture->setMonitorRotation();
}
void MonitorPictureProxy::updatePosition()
{
    monitorPicture->updatePos();
}
void MonitorPictureProxy::upatePrimary()
{
    monitorPicture->setPrimary();
}

MonitorPictureWidget::MonitorPictureWidget(KScreen::ConfigPtr config, QWidget * parent, Qt::WindowFlags f) :
    QWidget(parent,f)
{
    mConfig = config;
    mOutputs = mConfig->outputs();
    monitorsWidth = monitorsHeight = 0;
    ui.setupUi(this);
    setScene();

    ui.graphicsView->scale(0.1,0.1);
}
MonitorPictureWidget::~MonitorPictureWidget()
{
    scene->deleteLater();
}
void MonitorPictureWidget::addOutput(int key, KScreen::OutputPtr output)
{
    monitorsWidth += output->size().width();
    monitorsHeight += output->size().height();
    MonitorPicture *monitorPicture = new MonitorPicture(key, this);
    if(output->isEnabled())
        monitorPicture->setZValue(1);
    pictures.append(monitorPicture);
    scene->addItem(monitorPicture);
    MonitorPictureProxy *proxy = new MonitorPictureProxy(this, monitorPicture);
    proxy->connect(mOutputs[key].data(), SIGNAL(currentModeIdChanged()), SLOT(updatePosition()));
    proxy->connect(mOutputs[key].data(), SIGNAL(posChanged()), SLOT(updatePosition()));
    proxy->connect(mOutputs[key].data(), SIGNAL(rotationChanged()), SLOT(updateRotation()));
    proxy->connect(mOutputs[key].data(), SIGNAL(isPrimaryChanged()), SLOT(upatePrimary()));
    proxy->connect(mOutputs[key].data(), SIGNAL(isEnabledChanged()), SLOT(upatePrimary()));
}
void MonitorPictureWidget::outputChanged()
{
    for (KScreen::OutputPtr output : mOutputs) {
        bool isfind = false;
        if(!output->isConnected())
            continue;
        for (MonitorPicture * picture: pictures) {
            if(picture->mKey == output->id())
            {
                isfind = true;
                break;
            }
        }
        if(!isfind)
            addOutput(output->id(), output);
    }
}
void MonitorPictureWidget::setScene()
{
    scene = new QGraphicsScene(-ui.graphicsView->width()/2, -ui.graphicsView->height()/2,
                                               ui.graphicsView->width()/2, ui.graphicsView->height()/2);
    monitorsWidth = 0;
    monitorsHeight = 0;
    for (auto i = mOutputs.begin(); i != mOutputs.end(); ++i) {
        int key = i.key();
        KScreen::OutputPtr output = i.value();
        if(!output->isConnected())
            continue;
        addOutput(key, output);
    }
    connect(KScreen::ConfigMonitor::instance(), &KScreen::ConfigMonitor::configurationChanged,
            this, &MonitorPictureWidget::outputChanged);
    std::sort(pictures.begin(), pictures.end(), comparePositon);
    for (MonitorPicture * picture : pictures) {
        picture->updatePos();
    }
    ui.graphicsView->setScene(scene);
    ui.graphicsView->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    ui.graphicsView->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
}
MonitorPicture::MonitorPicture(int key,
                               MonitorPictureWidget *monitorPictureWidget) :
    QGraphicsRectItem(nullptr),
    mKey(key),
    monitorWidget(monitorPictureWidget)
{
    KScreen::OutputPtr &output = monitorPictureWidget->mOutputs[mKey];
    setAcceptedMouseButtons(Qt::LeftButton);
    setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemSendsGeometryChanges);
    setPen(QPen(QColor("#FFFFFF")));
    setBrush(output->isEnabled() ? output->isPrimary() ? QBrush(QColor("#2c9dfc")) : QBrush(QColor("#3cb034")) : QBrush(QColor("#011222")));

    textItem = new QGraphicsTextItem(output->name(), this);
    textItem->setScale(10);
    textItem->setDefaultTextColor(Qt::white);
    textItem->setParentItem(this);
}

void MonitorPicture::setPrimary()
{
    KScreen::OutputPtr output = monitorWidget->mOutputs[mKey];
    if(output->isEnabled())
        setZValue(1);
    else
        setZValue(0);
    setBrush(output->isEnabled() ? output->isPrimary() ? QBrush(QColor("#2c9dfc")) : QBrush(QColor("#3cb034")) : QBrush(QColor("#011222")));
}

void MonitorPicture::setMonitorRotation()
{
//    KScreen::OutputPtr output = monitorWidget->mOutputs[mKey];
//    KScreen::Output::Rotation rotation =  output->rotation();
//    QTransform transform;
//    transform.rotate(90, Qt::XAxis);
//    setTransform(transform);
//    textItem->setTransform(transform);
//    monitorWidget->update();
}
void MonitorPicture::updatePos()
{
    KScreen::OutputPtr output = monitorWidget->mOutputs[mKey];
    originX = output->pos().x();
    originY = output->pos().y();

    QList <KScreen::ModePtr> modeList = output->modes().values();
    if(modeList.size() <= 0)
        return;
    qSort(modeList.begin(), modeList.end(), sizeBiggerThan);
    QSize maxSize = modeList[0]->size();

    QSize currentSize;
    if(output->currentMode().isNull())
        currentSize = maxSize;
    else
        currentSize = output->currentMode()->size();

    if( output->rotation() == KScreen::Output::Left || output->rotation() == KScreen::Output::Right )
        currentSize.transpose();

//    scale = monitorWidget->height() / qMax(monitorWidget->monitorsWidth, monitorWidget->monitorsHeight);
//    setRect(0, 0, currentSize.width() * scale, currentSize.height() * scale);
//    setPos((originX  - monitorWidget->monitorsWidth /2) * scale, (originY - monitorWidget->monitorsHeight/2) * scale);

    setRect(0, 0, currentSize.width(), currentSize.height());
    setPos((originX  - monitorWidget->monitorsWidth /2), (originY - monitorWidget->monitorsHeight/2));
    qreal fontWidth = QFontMetrics(textItem->font()).width(output->name());
    qreal fontHeight = QFontMetrics(textItem->font()).height();
    textItem->setX((rect().width() - fontWidth*textItem->scale()) /2);
    textItem->setY((rect().height() - fontHeight*textItem->scale()) /2);
}
struct Parameters {
    float t1, t2;
    QVector2D cutPoint;
};

static Parameters segmentsCut(QVector2D p0, QVector2D p1, QVector2D s0, QVector2D s1)
{
    Parameters result;
    QVector2D v0 = p1 - p0;
    QVector2D v1 = s1 - s0;
    QVector2D P = s0 - p0;

    float det = v0.y() * v1.x() - v0.x() * v1.y();
    if (det == 0.0)
        result.t1 = result.t2 = -1.0;

    result.t1 = 1 / det * (-v1.y() * P.x() + v1.x() * P.y());
    result.t2 = 1 / det * (-v0.y() * P.x() + v0.x() * P.y());
    result.cutPoint = v0 * result.t1 + p0;
    return result;
}

static QVector2D computeCenter(MonitorPicture* monitorPicture)
{
    float x0 = monitorPicture->x();
    float y0 = monitorPicture->y();
    float x1 = x0 + monitorPicture->rect().width();
    float y1 = y0 + monitorPicture->rect().height();
    QVector2D p0(x0, y0);
    QVector2D p1(x1, y1);
    QVector2D center = p0 + (p1 - p0) * 0.5;
    return center;
}

struct Result_moveMonitorPictureToNearest
{
    bool ok;
    QVector2D vector;
};

static Result_moveMonitorPictureToNearest compareTwoMonitors(MonitorPicture* monitorPicture1,
                                                             MonitorPicture* monitorPicture2)
{
    Result_moveMonitorPictureToNearest result;
    QVector2D center1 = computeCenter(monitorPicture1);
    QVector2D center2 = computeCenter(monitorPicture2);
    float x0 = monitorPicture2->x();
    float y0 = monitorPicture2->y();
    float x1 = x0 + monitorPicture2->rect().width();
    float y1 = y0 + monitorPicture2->rect().height();

    QVector2D p0(x0, y0);
    QVector2D p1(x1, y1);
    QVector2D P1, P2;
    float t1 = -1.0, t2 = -1.0;

    Parameters params = segmentsCut(center1, center2, QVector2D(x0, y0), QVector2D(x1, y0));
    if (params.t1 >= 0.0 && params.t1 <= 1.0 && params.t2 >= 0.0 && params.t2 <= 1.0 && t1 < 0)
    {
        t1 = params.t1;
        P1 = params.cutPoint;
    }

    params = segmentsCut(center1, center2, QVector2D(x0, y0), QVector2D(x0, y1));
    if (params.t1 >= 0.0 && params.t1 <= 1.0 && params.t2 >= 0.0 && params.t2 <= 1.0 && t1 < 0)
    {
        t1 = params.t1;
        P1 = params.cutPoint;
    }

    params = segmentsCut(center1, center2, QVector2D(x1, y1), QVector2D(x1, y0));
    if (params.t1 >= 0.0 && params.t1 <= 1.0 && params.t2 >= 0.0 && params.t2 <= 1.0 && t1 < 0)
    {
        t1 = params.t1;
        P1 = params.cutPoint;
    }

    params = segmentsCut(center1, center2, QVector2D(x1, y1), QVector2D(x0, y1));
    if (params.t1 >= 0.0 && params.t1 <= 1.0 && params.t2 >= 0.0 && params.t2 <= 1.0 && t1 < 0)
    {
        t1 = params.t1;
        P1 = params.cutPoint;
    }

    x0 = monitorPicture1->x();
    y0 = monitorPicture1->y();
    x1 = x0 + monitorPicture1->rect().width();
    y1 = y0 + monitorPicture1->rect().height();
    p0 = QVector2D(x0, y0);
    p1 = QVector2D(x1, y1);

    params = segmentsCut(center1, center2, QVector2D(x0, y0), QVector2D(x1, y0));
    if (params.t1 >= 0.0 && params.t1 <= 1.0 && params.t2 >= 0.0 && params.t2 <= 1.0 && t2 < 0)
    {
        t2 = params.t1;
        P2 = params.cutPoint;
    }

    params = segmentsCut(center1, center2, QVector2D(x0, y0), QVector2D(x0, y1));
    if (params.t1 >= 0.0 && params.t1 <= 1.0 && params.t2 >= 0.0 && params.t2 <= 1.0 && t2 < 0)
    {
        t2 = params.t1;
        P2 = params.cutPoint;
    }

    params = segmentsCut(center1, center2, QVector2D(x1, y1), QVector2D(x1, y0));
    if (params.t1 >= 0.0 && params.t1 <= 1.0 && params.t2 >= 0.0 && params.t2 <= 1.0 && t2 < 0)
    {
        t2 = params.t1;
        P2 = params.cutPoint;
    }

    params = segmentsCut(center1, center2, QVector2D(x1, y1), QVector2D(x0, y1));
    if (params.t1 >= 0.0 && params.t1 <= 1.0 && params.t2 >= 0.0 && params.t2 <= 1.0 && t2 < 0)
    {
        t2 = params.t1;
        P2 = params.cutPoint;
    }

    // Monitor outside
//    if (t1 > t2)
//    {
//        result.vector = P1 - P2;
//        result.ok = false;
//    } else
//        result.ok = true;

    // Monitor outside
    if (t1 > t2)
    {
        result.vector = P1 - P2;
        result.ok = false;
    }
    // Monitor inside
    else
    {
        result.vector = P1 - P2;
        result.ok = false;
    }

    return result;
}

void MonitorPicture::mouseReleaseEvent(QGraphicsSceneMouseEvent * event)
{
    QGraphicsRectItem::mouseReleaseEvent(event);
    monitorWidget->moveMonitorPictureToCenter(this);
    monitorWidget->updateMonitorWidgets();
}
void MonitorPictureWidget::updateMonitorWidgets()
{
    for (MonitorPicture *picture : qAsConst(pictures))
    {
//        int x = monitorsWidth /2 + picture->pos().x() / picture->scale;
//        int y = monitorsHeight /2 + picture->pos().y() / picture->scale;
        int x = monitorsWidth /2 + picture->pos().x();
        int y = monitorsHeight /2 + picture->pos().y();
        KScreen::OutputPtr output = mOutputs[picture->mKey];
        output->setPos(QPoint(x, y));
        if(x > 0 || y > 0) {
            emit positionChanged();
        }
    }
}
void MonitorPictureWidget::moveMonitorPictureToCenter(MonitorPicture *monitorPicture)
{
    QVector2D vector(0, 0);
    for (MonitorPicture *picture : qAsConst(pictures))
    {
        if (picture == monitorPicture)
            continue;

        Result_moveMonitorPictureToNearest result = compareTwoMonitors(monitorPicture, picture);
        if (result.ok)
            return;
        else if (result.vector.length() < vector.length() || vector.length() == 0.0)
            vector = result.vector;
    }

    int x = monitorPicture->x();
    int y = monitorPicture->y();
    monitorPicture->setX(x + vector.x());
    monitorPicture->setY(y + vector.y());
}
