#include "screenmonitor.h"
#include "ui_screenmonitor.h"
#include "monitor.h"
#include "monitoroutputwidget.h"
#include <KScreen/ConfigMonitor>
#include <KScreen/SetConfigOperation>
#include <KScreen/Screen>
#include <KScreen/Output>
#include <KScreen/EDID>
#include <Nde/Settings>
#include <Nde/AutostartEntry>
#include <QDateTime>
#include <QPushButton>
#include "../baseui/common.h"
#include "timeoutdialog.h"

ScreenMonitor::ScreenMonitor(QWidget *parent) :
    Page(parent),
    ui(new Ui::ScreenMonitor)
{
    ui->setupUi(this);
    KScreen::GetConfigOperation *operation = new KScreen::GetConfigOperation();

    connect(operation, &KScreen::GetConfigOperation::finished, [this, operation] (KScreen::ConfigOperation *op) {
        KScreen::GetConfigOperation *configOp = qobject_cast<KScreen::GetConfigOperation *>(op);
        if (configOp && configOp->config() && configOp->config()->screen())
        {
            mOldConfig = configOp->config()->clone();
            loadConfiguration(configOp->config());
            operation->deleteLater();
        }
        else if(configOp && !configOp->config())
        {
            qDebug() << "Error: Config is invalid, probably backend couldn't load";
            exit(1);
        }
        else if(configOp && configOp->config() && !configOp->config()->screen())
        {
            qDebug() << "Error: No screen in the configuration, broken backend";
            exit(2);
        }
        else
        {
            qDebug() << "Error: Connect to KScreen is not possible";
            exit(3);
        }
    });
}
ScreenMonitor::~ScreenMonitor()
{
    delete ui;
}

void ScreenMonitor::applyBtnClicked()
{
    if (mConfig && KScreen::Config::canBeApplied(mConfig))
    {
        KScreen::SetConfigOperation(mConfig).exec();

        TimeoutDialog mTimeoutDialog;
        if (mTimeoutDialog.exec() == QDialog::Rejected) {
            KScreen::SetConfigOperation(mOldConfig).exec();
        }
        else
        {
            mOldConfig = mConfig->clone();
            saveConfiguration(mConfig);
        }
    }
}

void ScreenMonitor::refreshPage()
{

}

void ScreenMonitor::saveConfiguration(KScreen::ConfigPtr config)
{
    QList<MonitorSettings> currentSettings;
    KScreen::OutputList outputs = config->outputs();
    for (const KScreen::OutputPtr &output : outputs)
    {
        MonitorSettings monitor;
        monitor.name = output->name();
        KScreen::Edid* edid = output->edid();
        if (edid && edid->isValid())
            monitor.hash = edid->hash();
        monitor.connected = output->isConnected();
        monitor.enabled = output->isEnabled();
        
        if ( output->isConnected() && output->isEnabled() )
        {
            
            monitor.primary = output->isPrimary();
            monitor.xPos = output->pos().x();
            monitor.yPos = output->pos().y();
            monitor.currentMode = output->currentModeId();
            monitor.currentModeWidth = output->currentMode()->size().width();
            monitor.currentModeHeight = output->currentMode()->size().height();
            monitor.currentModeRate = output->currentMode()->refreshRate();
            monitor.rotation = output->rotation();
        }
        currentSettings.append(monitor);
    }

    Nde::Settings settings("nde-systemsettings-monitor");
    settings.beginGroup("currentConfig");
    saveMonitorSettings(settings, currentSettings);
    settings.endGroup();

    QList<MonitorSavedSettings> monitors;
    settings.beginGroup("SavedConfigs");
    loadMonitorSettings(settings, monitors);
    qDebug() << "[ MonitorSettingsDialog::saveConfiguration] # monitors Read:" << monitors.size();
    MonitorSavedSettings monitor;
    monitor.name = QDateTime::currentDateTime().toString();
    monitor.date = QDateTime::currentDateTime().toString(Qt::ISODate);
    monitor.monitors = currentSettings;
    monitors.append(monitor);
    saveMonitorSettings(settings, monitors);
    qDebug() << "[ MonitorSettingsDialog::saveConfiguration] # monitors Write:" << monitors.size();
    settings.endGroup();

    Nde::AutostartEntry autoStart("nde-systemsettings-autostart.desktop");
    XdgDesktopFile desktopFile(XdgDesktopFile::ApplicationType, "nde-systemsettings-autostart", "nde-systemsettings-autostart");
    desktopFile.setValue("OnlyShowIn", "Nde");
    desktopFile.setValue("Comment", "Autostart monitor settings for nde-systemsettings");
    autoStart.setFile(desktopFile);
    autoStart.commit();
}
void ScreenMonitor::loadConfiguration(KScreen::ConfigPtr config)
{
    if (mConfig == config)
        return;

    mConfig = config;
    KScreen::ConfigMonitor::instance()->addConfig(mConfig);
    MonitorOutputWidget *monitor = new MonitorOutputWidget(config, this);
    mMonitorPictureWidget = new MonitorPictureWidget(config, this);
    connect(mMonitorPictureWidget, &MonitorPictureWidget::positionChanged, monitor, &MonitorOutputWidget::positionChanged);
    ui->displayVBoxLayout->addWidget(mMonitorPictureWidget);
    ui->displayVBoxLayout->addWidget(monitor);
}
