#include <QX11Info>
#include <QDebug>

#include "xrandrbrightness.h"

XRandrBrightness::XRandrBrightness()
{
    xcb_screen_iterator_t iter = xcb_setup_roots_iterator(xcb_get_setup(QX11Info::connection()));
    if (!iter.rem) {
        qDebug() << "XCB Screen Roots Iterator rem was null";
        return;
    }

    xcb_screen_t *screen = iter.data;
    xcb_window_t root = screen->root;

    m_resources.reset(xcb_randr_get_screen_resources_current_reply(QX11Info::connection(),
        xcb_randr_get_screen_resources_current(QX11Info::connection(), root)
    , nullptr));

    if (!m_resources) {
        qDebug() << "RANDR Get Screen Resources returned null";
        return;
    }
}
bool XRandrBrightness::canBrightness()
{
    if (!QX11Info::isPlatformX11()) {
        return false;
    }
    ScopedCPointer<xcb_randr_query_version_reply_t> versionReply(xcb_randr_query_version_reply(QX11Info::connection(),
        xcb_randr_query_version(QX11Info::connection(), 1, 2),
    nullptr));

    if (!versionReply) {
        qDebug() << "RandR Query version returned null";
        return false;
    }

    if (versionReply->major_version < 1 || (versionReply->major_version == 1 && versionReply->minor_version < 2)) {
        qDebug() << "RandR version" << versionReply->major_version << "." << versionReply->minor_version << " too old";
        return false;
    }
    return true;
}
float XRandrBrightness::gamma_brightness_get()
{
    if (!m_resources) {
        return -1;
    }

    auto *outputs = xcb_randr_get_screen_resources_current_outputs(m_resources.data());
    for (int i = 0; i < m_resources->num_outputs; ++i) {

        xcb_randr_output_t output = outputs[i];
        xcb_generic_error_t  *error;
        xcb_randr_get_output_info_cookie_t output_info_cookie = xcb_randr_get_output_info (QX11Info::connection(), output, 0);
        xcb_randr_get_output_info_reply_t * output_info = xcb_randr_get_output_info_reply (QX11Info::connection(), output_info_cookie, &error);
        if(error != nullptr)
        {
            qDebug() << "Error getting output_info";
            return -1;
        }
        if(output_info == nullptr)
        {
            qDebug() << "Error: output_info is null";
            return -1;
        }
        QString name = QString::fromUtf8((const char * ) xcb_randr_get_output_info_name(output_info), output_info->name_len);

        //Is Connection?
        if ( static_cast<xcb_randr_connection_t>(output_info->connection) != XCB_RANDR_CONNECTION_CONNECTED )
            continue;

        // Is enabled?
        if( output_info->crtc == 0)
            continue;
        if(kScreenOutput->name() != name || kScreenOutput->id() != static_cast<int>(output))
            continue;
        // xcb_randr_get_output_info_reply_t tiene como elemento crtc
        xcb_randr_get_crtc_gamma_cookie_t gamma_cookie = xcb_randr_get_crtc_gamma_unchecked (QX11Info::connection(), output_info->crtc);
        xcb_randr_get_crtc_gamma_reply_t *gamma_reply =  xcb_randr_get_crtc_gamma_reply (QX11Info::connection(), gamma_cookie, &error);
        if(error != nullptr)
        {
            qDebug() << "Error getting gamma_reply";
            return -1;
        }
        if(gamma_reply == nullptr)
        {
            qDebug() << "Error: gamma_reply is null";
            return -1;
        }
        uint16_t *red = xcb_randr_get_crtc_gamma_red (gamma_reply);
        if(red == nullptr)
        {
            qDebug() << "Error: red is null";
            return -1;
        }
        int red_length = xcb_randr_get_crtc_gamma_red_length(gamma_reply);
        float brightness = static_cast<float>(red[red_length-1]/65535.0);
        return brightness;
    }
    return -1;
}

void XRandrBrightness::gamma_brightness_set(float percent)
{
    if (!m_resources) {
        return;
    }
    auto *outputs = xcb_randr_get_screen_resources_current_outputs(m_resources.data());
    for (int i = 0; i < m_resources->num_outputs; ++i) {
        xcb_generic_error_t  *error;
        xcb_randr_output_t output = outputs[i];
        xcb_randr_get_output_info_cookie_t output_info_cookie = xcb_randr_get_output_info (QX11Info::connection(), output, 0);
        xcb_randr_get_output_info_reply_t * output_info = xcb_randr_get_output_info_reply (QX11Info::connection(), output_info_cookie, &error);
        if(error != nullptr)
        {
            qDebug() << "Error getting output_info";
            return;
        }
        if(output_info == nullptr)
        {
            qDebug() << "Error: output_info is null";
            return;
        }
        QString name = QString::fromUtf8((const char * ) xcb_randr_get_output_info_name(output_info), output_info->name_len);
        //Is Connection?
        if ( static_cast<xcb_randr_connection_t>(output_info->connection) != XCB_RANDR_CONNECTION_CONNECTED )
            continue;

        // Is enabled?
        if( output_info->crtc == 0)
            continue;
        if(kScreenOutput->name() != name || kScreenOutput->id() != static_cast<int>(output))
            continue;
        // xcb_randr_get_output_info_reply_t tiene como elemento crtc
        xcb_randr_get_crtc_gamma_cookie_t gamma_cookie = xcb_randr_get_crtc_gamma_unchecked (QX11Info::connection(), output_info->crtc);
        xcb_randr_get_crtc_gamma_reply_t *gamma_reply =  xcb_randr_get_crtc_gamma_reply (QX11Info::connection(), gamma_cookie, &error);
        if(error != nullptr)
        {
            qDebug() << "Error getting gamma_reply";
            return;
        }
        if(gamma_reply == nullptr)
        {
            qDebug() << "Error: gamma_reply is null";
            return;
        }
        uint16_t *red = xcb_randr_get_crtc_gamma_red (gamma_reply);
        if(red == nullptr)
        {
            qDebug() << "Error: red is null";
            return;
        }
        int red_length = xcb_randr_get_crtc_gamma_red_length(gamma_reply);
        uint16_t *green = xcb_randr_get_crtc_gamma_green (gamma_reply);
        if(green == nullptr)
        {
            qDebug() << "Error: green is null";
            return;
        }
        uint16_t *blue = xcb_randr_get_crtc_gamma_blue (gamma_reply);
        if(blue == nullptr)
        {
            qDebug() << "Error: blue is null";
            return;
        }

        float max_gamma = 65535*percent;
        for(int i=0;i<red_length;i++)
        {
            int value = qMin((int)(((float)i/(float)(red_length-1))*max_gamma),65535);
            green[i] = blue[i] = red[i] = value;
        }
        xcb_randr_set_crtc_gamma (QX11Info::connection(), output_info->crtc, red_length, red, green, blue);
    }
    return;
}
void XRandrBrightness::setOutput(const KScreen::OutputPtr &output)
{
    kScreenOutput = output;
}
