#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <XdgDesktopFile>
#include <XdgMenu>
#include <QMessageBox>
#include <XmlHelper>
#include <XdgIcon>
#include <QStyledItemDelegate>
#include <QPainter>
#include <QPluginLoader>
#include "settingspluginlibrary.h"
#include <XdgDirs>
#include <QDebug>
#include <QDir>
#include <QProcessEnvironment>
#include <QScrollBar>
#include <QApplication>
#include <QScreen>
#include <QTimer>
#include <QTime>
#include <KWindowSystem/KWindowSystem>
#include "categorizedmodelitem.h"

#define MIN_SIZE QSize(1024,600)

extern void *loadPluginTranslation_systeminfo_helper;
QStringList pluginDesktopDirs()
{
    QStringList dirs;
    dirs << QString(getenv("NDE_PANEL_PLUGINS_DIR")).split(':', QString::SkipEmptyParts);
    dirs << QString("%1/%2").arg(XdgDirs::dataHome(), "/nde/nde-systemsettings");
    dirs << PLUGIN_DESKTOPS_DIR;
    qDebug () << QString("%1/%2").arg(XdgDirs::dataHome(), "/nde/nde-systemsettings") << PLUGIN_DESKTOPS_DIR;
    return dirs;
}
MainWindow::MainWindow(const QString module, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    model_ = new CategorizedModel (pluginDesktopDirs(), ui->categoryWidget);
    model_->initModel();
//    connect(model_, &CategorizedModel::childItemAdded, this, &MainWindow::addItem);
    connect(ui->categoryWidget, &CategorizedTreeView::clicked, this, &MainWindow::itemClicked);
    ui->categoryWidget->setModel(model_);
    ui->categoryWidget->expandAll();
    ui->sidebarWidget->setStyleSheet("QWidget{background:#f0f1f2;}");

    connect(qApp->primaryScreen(),  &QScreen::availableGeometryChanged, this, &MainWindow::onScreenGeometryChanged);
    onScreenGeometryChanged(qApp->primaryScreen()->availableGeometry());

    //设置默认显示的页面模块
    QTimer::singleShot(200, [this, module](){
        setCurrentItem(module);
    });
}

void MainWindow::wshow()
{
    KWindowSystem::forceActiveWindow(this->winId());
}

void MainWindow::itemClicked(const QModelIndex &index)
{
    CategorizedModelItem *modelItem = static_cast<CategorizedModelItem *>(model_->itemFromIndex(index));
    QString id = modelItem->id();
    setCurrentItem(id);
}

void MainWindow::onScreenGeometryChanged(const QRect &geometry)
{
    QSize min = MIN_SIZE;
    if(geometry.size().width() < MIN_SIZE.width())
    {
        min.rwidth() = geometry.size().width();
    }
    if(geometry.size().height() - 40 < MIN_SIZE.height())
    {
        min.rheight() = geometry.size().height() - 40;
    }

    this->setMinimumSize(min);
    this->update();
}

void MainWindow::addItem(const QString &id,const int &status)
{
    QStringList dirs;
    dirs << QProcessEnvironment::systemEnvironment().value("NDEPANEL_PLUGIN_PATH").split(":");
    dirs << PLUGIN_DIR;

    QString baseName = QString("lib%1.so").arg(id);
    for(const QString &dirName : qAsConst(dirs))
    {
        QFileInfo fi(QDir(dirName), baseName);
        if (fi.exists())
        {
            if (loadPage(id , fi.absoluteFilePath(),status))
                break;
        }
    }
}
bool MainWindow::loadPage(const QString &id, const QString &libraryName,const int &status)
{
    QPluginLoader pluginLoader(libraryName);

    if (!pluginLoader.load())
    {
        qWarning() << pluginLoader.errorString();
        return false;
    }

    QObject *obj = pluginLoader.instance();
    if (!obj)
    {
        qWarning() << pluginLoader.errorString();
        return false;
    }

    SettingsPluginLibrary* pluginLib= qobject_cast<SettingsPluginLibrary*>(obj);
    if (!pluginLib)
    {
        qWarning() << QString("Can't load plugin \"%1\". Plugin is not a SettingsPluginLibrary.").arg(id);
        delete obj;
        return false;
    }

//    QTime t = QTime::currentTime();
    StackPage * sPage = pluginLib->create(this);
    if(!status)
    {
        sPage->setDisabled(true); //设置右侧界面显示状态
    }
    sPage->hide();
    mPluginPageHash.insert(id, sPage);
//    QTime t1 = QTime::currentTime();
//    qWarning() << id << "---------------" << t.msecsTo(t1);
    return true;
}

void MainWindow::setCurrentItem(const QString &id)
{
    if(!id.isEmpty() && mPluginPageHash.find(id) == mPluginPageHash.end())
    {
        addItem(id, model_->pluginEnable(id));
    }

    StackPage * sPage = nullptr;
    if(!id.isEmpty() && mPluginPageHash.find(id) != mPluginPageHash.end())
    {
        sPage = mPluginPageHash.value(id);
    }
//    else if(!mPluginPageHash.isEmpty()) {
//        baseWidget = mPluginPageHash.begin().value();
//    }

    if(sPage)
    {
        //先取出之前widget。若不取出，直接ui->scrollArea->setWidget(sss)设置将会导致之前的widget被删除
        ui->scrollArea->takeWidget();
        ui->scrollArea->setWidget(sPage);
        sPage->show();
        sPage->refreshStackPage();
        ui->categoryWidget->selectionModel()->select(model_->indexFromItem(model_->SelectItem(id)), QItemSelectionModel::ClearAndSelect);
    }
}

MainWindow::~MainWindow()
{
    delete ui;
}
