#include "systeminfomanager.h"
#include <QFile>
#include <QDebug>
#include <QSysInfo>
#include <QStorageInfo>
#include <QSettings>
#include <QProcess>

SystemInfoManager::SystemInfoManager(QObject *parent) : QObject(parent)
{
    onSysInfoChanged();
    onCPUChanged();
    onRAMChanged();
    onOsReleaseChanged();
}

SystemInfoManager *SystemInfoManager::getIns()
{
    static SystemInfoManager mgr;
    return &mgr;
}

QString SystemInfoManager::getSystemInfo(const QString type)
{
    QString info ;
    if(mSystemInfo.contains(type))
        info = mSystemInfo.value(type);
    return info;
}

QStringList SystemInfoManager::getDisksInfo()
{
    QStringList infos;
    return infos;
}

void SystemInfoManager::onSysInfoChanged()
{
    //cpu架构 x86_64 i686
    mSystemInfo["arch"] = QSysInfo::currentCpuArchitecture();
    //linux 4.19.112-2.el8.X86_64
    mSystemInfo["kernel"] = QSysInfo::kernelType() + " " + QSysInfo::kernelVersion();
    //---
    mSystemInfo["hostName"] = QSysInfo::machineHostName();

    QStorageInfo storage = QStorageInfo::root();
    //系统容量
    mSystemInfo["rootSize"] = QString::number(storage.bytesTotal()/(1024 * 1024 * 1024 * 1.0), 'f', 1) + "GB";
    qint64 root_free = storage.bytesFree() > storage.bytesAvailable() ? storage.bytesAvailable() : storage.bytesFree();
    mSystemInfo["rootFree"] = QString::number(root_free/(1024 * 1024 * 1024 * 1.0), 'f', 1) + "GB";

    //用户目录容量
    QList<QStorageInfo> lists = QStorageInfo::mountedVolumes();
    for(int i = 0; i < lists.count(); i++)
    {
        QStorageInfo info = lists.at(i);
        if(info.displayName() == "/home"){
            mSystemInfo["homeSize"] = QString::number(info.bytesTotal()/(1024 * 1024 * 1024 * 1.0), 'f', 1) + "GB";
            qint64 free = info.bytesFree() > info.bytesAvailable() ? info.bytesAvailable() : info.bytesFree();
            mSystemInfo["homeFree"] = QString::number(free/(1024 * 1024 * 1024 * 1.0), 'f', 1) + "GB";
            break;
        }
    }
}

void SystemInfoManager::onOsReleaseChanged()
{
    QSettings setting("/etc/os-release", QSettings::IniFormat, this);
    mSystemInfo["productName"] = setting.value("NAME").toString();
    QString version = setting.value("VERSION").toString();
    QString buildID = setting.value("BUILD_ID").toString();
    QString fullVersion;
    if(version.contains("-"))
    {
        QStringList versoinList = version.split("-");
        QString v = versoinList.takeFirst();
        QString s = versoinList.takeFirst();

        fullVersion = v + "." + buildID + "-" + s;
    }
    else
    {
        fullVersion = version + "." + buildID;
    }
    mSystemInfo["productVersion"] = version;
    mSystemInfo["productFullVersion"] = fullVersion;
    mSystemInfo["productPrettyName"] = setting.value("PRETTY_NAME").toString();
}

void SystemInfoManager::onCPUChanged()
{
    QString cpuCores;
    QString modelName;
    QFile file("/proc/cpuinfo");
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        return;
    }
    QTextStream in(&file);
    QString line = in.readLine();
    while (!in.atEnd()) {

        QStringList lineList = line.split(":");
        if(lineList.at(0).trimmed() == "cpu cores")
        {
            cpuCores = lineList.at(1).trimmed();
        }
        if(lineList.at(0).trimmed() == "model name")
        {
            modelName = lineList.at(1).trimmed();
        }
        if(!modelName.isEmpty() && !cpuCores.isEmpty())
        {
            mSystemInfo["cpuInfo"]  = modelName + " x " + cpuCores;
            break;
        }
        line = in.readLine();
    }
    file.close();

    if(!mSystemInfo.contains("cpuInfo")){
        QString cmd = "sudo dmidecode -t processor";
        QProcess process;
        process.start(cmd);
        process.waitForFinished();
        QString result = process.readAllStandardOutput();
        if(!result.isEmpty()){
            QStringList list = result.split(QRegExp("[:\t\r\n]"), QString::SkipEmptyParts);
            if(modelName.isEmpty())
                modelName = list.at(list.indexOf("Version")+1);
            if(cpuCores.isEmpty())
                cpuCores = list.at(list.indexOf("Core Count")+1);
            if(!modelName.isEmpty() && !cpuCores.isEmpty())
                mSystemInfo["cpuInfo"]  = modelName + " x " + cpuCores;
        }
    }
}

void SystemInfoManager::onRAMChanged()
{
    QFile file("/proc/meminfo");
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        return;
    }
    QTextStream in(&file);
    QString line = in.readLine();
    while (!in.atEnd()) {
        QStringList lineList = line.split(":");
        if(lineList.at(0).trimmed() == "MemTotal")
        {
            QString memTotalByKb = lineList.at(1).trimmed();
            double memSizeByGiB = memTotalByKb.split(" ").at(0).toDouble() / (1024 * 1024);//GiB/GB???
            mSystemInfo["meminfo"] = QString::number(memSizeByGiB, 'f', 1) + "GB";
            break;
        }
        line = in.readLine();
    }
    file.close();
}
