#include "wifilistwidgetitem.h"
#include "ui_wifilistwidgetitem.h"
#include <QDebug>
#include <QPainter>
#include "nmconfig.h"
#include <QDialog>
#include <QLineEdit>
#include <QMessageBox>
#include <QKeyEvent>
#include "../baseui/common.h"
#include <QTimer>

NetworkManager::ConnectionSettings::Ptr assembleWpaXPskSettings(const NetworkManager::AccessPoint::Ptr accessPoint, const QString &password, bool autoConnect)
{
    NetworkManager::ConnectionSettings::Ptr settings{new NetworkManager::ConnectionSettings{NetworkManager::ConnectionSettings::Wireless}};
    settings->setId(accessPoint->ssid());
    settings->setUuid(NetworkManager::ConnectionSettings::createNewUuid());
    settings->setAutoconnect(autoConnect);
    //Note: workaround for wrongly (randomly) initialized gateway-ping-timeout
    settings->setGatewayPingTimeout(0);

    NetworkManager::WirelessSetting::Ptr wifi_sett
        = settings->setting(NetworkManager::Setting::Wireless).dynamicCast<NetworkManager::WirelessSetting>();
    wifi_sett->setInitialized(true);
    wifi_sett->setSsid(accessPoint->ssid().toUtf8());
    wifi_sett->setSecurity("802-11-wireless-security");

    NetworkManager::WirelessSecuritySetting::Ptr security_sett
        = settings->setting(NetworkManager::Setting::WirelessSecurity).dynamicCast<NetworkManager::WirelessSecuritySetting>();
    security_sett->setInitialized(true);
    if (NetworkManager::AccessPoint::Adhoc == accessPoint->mode())
    {
        wifi_sett->setMode(NetworkManager::WirelessSetting::Adhoc);
        security_sett->setKeyMgmt(NetworkManager::WirelessSecuritySetting::WpaNone);
    } else
    {
        security_sett->setKeyMgmt(NetworkManager::WirelessSecuritySetting::WpaPsk);
    }
    security_sett->setPsk(password);
    return settings;
}

WifiListWidgetItem::WifiListWidgetItem(NetworkManager::WirelessDevice * device, NetworkManager::WirelessNetwork::Ptr network, QWidget *parent) :
    QWidget(parent),
    ui(new Ui::WifiListWidgetItem),
    m_device(device),
    m_network(network),
    m_ssid(network->ssid()),
    m_signalStrength(network->signalStrength()),
    mConnectingTimerCount(0),
    m_bClicked(false),
    m_showType(ShowStatus),
    m_bToDelete(false)
{
    ui->setupUi(this);

    QFile f(":/qss/wifilistwidgetitem.qss");
    if(f.open(QIODevice::ReadOnly))
    {
        QString style = f.readAll().replace("\n", "");
        this->setStyleSheet(style);
    }

    ui->widget_password->setVisible(false);
    ui->label_ssid->setText(m_ssid);
    ui->label_signal->setPixmap(QPixmap(getIconBySignalStrength(m_signalStrength)));
    ui->label_signal->setAlignment(Qt::AlignCenter);

    QString pass;
    if(checkNeedPassword(pass))
    {
        ui->label_pwd->setPixmap(QPixmap(":/image/pwd.png"));
    }
    mConnectingTimer = new QTimer(this);
    mConnectingTimer->setInterval(300);
    mConnectingTimer->setSingleShot(false);
    connect(mConnectingTimer, &QTimer::timeout, [this](){
        QString dotdotdot;
        dotdotdot.fill('.', mConnectingTimerCount);
        if(mConnectingTimerCount >=3) mConnectingTimerCount = 0;
        else mConnectingTimerCount++;
        ui->label_ssid->setText( m_ssid + "   " + tr("Connecting") + dotdotdot);
        checkActive();
    });

    connect(ui->btn_setting, &QPushButton::clicked, this, &WifiListWidgetItem::editClicked);
    connect(ui->btn_cancel, &QPushButton::clicked, this, &WifiListWidgetItem::onBtnCancelClicked);
    connect(ui->btn_connect, &QPushButton::clicked, this, &WifiListWidgetItem::onBtnConnectClicked);
    connect(ui->btn_disconnect, &QPushButton::clicked, this, &WifiListWidgetItem::onBtnDisconnectClicked);
    connect(ui->edit_password, &QLineEdit::textChanged, this, &WifiListWidgetItem::onPasswordTextChanged);
    connect(ui->edit_password, &QLineEdit::editingFinished, this, &WifiListWidgetItem::onPasswordTextChanged);

    connect(m_device, &NetworkManager::Device::stateChanged, this, [this](NetworkManager::Device::State newstate, NetworkManager::Device::State oldstate, NetworkManager::Device::StateChangeReason reason){
        if(mConnectingTimer->isActive())
        {
            if(newstate == NetworkManager::Device::State::Failed)
            {
                ui->label_msg->setVisible(true);
                //提示3秒隐藏
                QTimer::singleShot(3 * 1000, this, [this](){ui->label_msg->setVisible(false);});
            }
            else if(newstate == NetworkManager::Device::State::Activated){
                showStatus();
            }
        }
    });
    ui->label_msg->setVisible(false);

    ui->edit_password->installEventFilter(this);


    m_failMsgs << tr("Wlan")
         << static_cast<unsigned int>(0)
         << "icon-l-settings"
         << tr("Could not connect wifi")
         << QString("\"%1\"").arg(m_ssid)
         << QStringList()
         << QVariantMap()
         << 5000;
}

WifiListWidgetItem::~WifiListWidgetItem()
{
    delete ui;
}

NetworkManager::WirelessNetwork::Ptr WifiListWidgetItem::getNetwork() const
{
    return m_network;
}

NetworkManager::WirelessDevice *WifiListWidgetItem::getDevice() const
{
    return  m_device;
}

QString WifiListWidgetItem::getSecType() const
{
    switch (m_secType) {
    case NetworkManager::WpaPsk:
        return "WpaPsk";
    case NetworkManager::Wpa2Psk:
        return "Wpa2Psk";
    default:
        return "NoneSecurity";
    }
}

void WifiListWidgetItem::updateActiveStatu()
{
    checkActive();

    switch (m_showType) {
    case ShowInit:
        showCurrentLinkStatus(ShowStatus);
        break;
    case ShowLink:
        showCurrentLinkStatus(ShowStatus);
        break;
    case ShowStatus:
        showCurrentLinkStatus(ShowStatus);
        break;
    case ShowInputKey:
        showCurrentLinkStatus(ShowStatus);
        break;
    case ShowLinking:
    {
        if(m_active == NetworkManager::ActiveConnection::Activating)
            showCurrentLinkStatus(ShowLinking);
        else if(m_active == NetworkManager::ActiveConnection::Activated)
            showCurrentLinkStatus(ShowLinkFinished);
        else
            showCurrentLinkStatus(ShowStatus);

        break;
    }
    case ShowLinkFinished:
        showCurrentLinkStatus(ShowStatus);
        break;
    }
}

bool WifiListWidgetItem::getActiveStatu()
{
    checkActive();
    return (m_active == NetworkManager::ActiveConnection::Activated || m_active == NetworkManager::ActiveConnection::Activating);
}

void WifiListWidgetItem::setEditVisible(bool b)
{
    ui->widget_password->setVisible(b);
    if(b)
    {
        checkActive();
        if(m_active == NetworkManager::ActiveConnection::Activated)
        {
            ui->horizontalWidget_2->setVisible(false);
            ui->frame_connect->setVisible(false);
            ui->frame_disconnect->setVisible(true);
        }
        else {
            ui->horizontalWidget_2->setVisible(true);
            ui->frame_connect->setVisible(true);
            ui->frame_disconnect->setVisible(false);
            ui->edit_password->setFocus();
        }
    }
}

bool WifiListWidgetItem::getEditVisible()
{
    return ui->widget_password->isVisible();
}

void WifiListWidgetItem::setEditFocus(bool b)
{
    if(b)
        ui->edit_password->setFocus();
    else
        ui->edit_password->clearFocus();
}

bool WifiListWidgetItem::getEditFocus()
{
    return ui->edit_password->hasFocus();
}

std::pair<QString, int> WifiListWidgetItem::getEditText()
{
    return std::make_pair<QString, int>(ui->edit_password->text(), ui->edit_password->cursorPosition());
}

void WifiListWidgetItem::setEditText(std::pair<QString, int> text)
{
    ui->edit_password->setText(text.first);
    ui->edit_password->setCursorPosition(text.second);
}

bool WifiListWidgetItem::getCheckBoxStatu()
{
    return ui->checkBox_password->isChecked();
}

void WifiListWidgetItem::setCheckBoxStatu(bool b)
{
    ui->checkBox_password->setChecked(b);
}

void WifiListWidgetItem::onSignalStrengthChanged(int signalStrength)
{
    ui->label_signal->setPixmap(QPixmap(getIconBySignalStrength(signalStrength)));
    ui->label_signal->setAlignment(Qt::AlignCenter);
    m_signalStrength = signalStrength;
}

void WifiListWidgetItem::onLinkForget()
{
    NetworkManager::Connection::Ptr conn = getConnect();
    if(conn && conn->isValid())
    {
        QString command = QString::fromUtf8("nmcli connect delete \"%1\"").arg(conn->name());
        bool res = QProcess::startDetached(command);
        if(res)
        {
            this->showCurrentLinkStatus(ShowLink);
            qWarning() <<  QString::fromUtf8("delete connect \"%1\" successed.").arg(conn->name());
        }

        //删除保存的密钥
        NMConfig::global()->removeWiFIPasswd(conn->name());
    }

    updateActiveStatu();
}

void WifiListWidgetItem::onLinkBreak()
{
    for(NetworkManager::ActiveConnection::Ptr aConn: NetworkManager::activeConnections())
    {
        if(aConn->id() == m_network->ssid())
        {
            QDBusPendingReply<> reply = NetworkManager::deactivateConnection(aConn->path());
            QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(reply, this);
            connect(watcher, &QDBusPendingCallWatcher::finished, [this,aConn] (QDBusPendingCallWatcher * watcher) {
                if (watcher->isError() || !watcher->isValid())
                {
                    qWarning() << QStringLiteral("断开无线连接['%1']失败: %2。").arg(aConn->id()).arg(watcher->error().message());
                }
                watcher->deleteLater();
            });
            break;
        }
    }

    updateActiveStatu();
}

void WifiListWidgetItem::onBtnCancelClicked()
{
    if(m_bToDelete)
    {
        Q_EMIT itemDelete();
    }
    ui->widget_password->setVisible(false);
}

void WifiListWidgetItem::onBtnConnectClicked()
{
    if(m_bToDelete)
    {
        NMConfig::global()->notifyMsg(m_failMsgs);
        Q_EMIT itemDelete();
        return;
    }
    Common::ChattrDir chat;

    connectByPwd(ui->edit_password->text());
    ui->widget_password->setVisible(false);
}

void WifiListWidgetItem::onBtnDisconnectClicked()
{
    onLinkBreak();
    ui->widget_password->setVisible(false);
}

void WifiListWidgetItem::onPasswordTextChanged()
{
    if(ui->edit_password->text().size() >= 8)
        ui->btn_connect->setEnabled(true);
    else {
        ui->btn_connect->setEnabled(false);
    }
}

void WifiListWidgetItem::mousePressEvent(QMouseEvent *event)
{
    m_bClicked = true;
    update();
}

void WifiListWidgetItem::mouseReleaseEvent(QMouseEvent *event)
{
    m_bClicked = false;
    if(m_needPassword)
    {
        ui->label_msg->setVisible(false);
        ui->edit_password->setText(NMConfig::global()->readWIFIPasswd(m_ssid));
        if(ui->edit_password->text().size() < 8)
            ui->btn_connect->setEnabled(false);
        ui->widget_password->setVisible(!ui->widget_password->isVisible());
        if(m_bToDelete && !ui->widget_password->isVisible())
        {
            Q_EMIT itemDelete();
            return;
        }
        if(ui->widget_password->isVisible())
        {
            checkActive();
            if(m_active == NetworkManager::ActiveConnection::Activated)
            {
                ui->horizontalWidget_2->setVisible(false);
                ui->frame_connect->setVisible(false);
                ui->frame_disconnect->setVisible(true);
            }
            else {
                ui->horizontalWidget_2->setVisible(true);
                ui->frame_connect->setVisible(true);
                ui->frame_disconnect->setVisible(false);
                ui->edit_password->setFocus();
            }
        }
    }
    else {
        if(m_bToDelete)
        {
            NMConfig::global()->notifyMsg(m_failMsgs);
            Q_EMIT itemDelete();
            return;
        }
        connectNoPwd();
    }
    update();
    Q_EMIT itemClicked();
}

void WifiListWidgetItem::paintEvent(QPaintEvent *event)
{
    Q_UNUSED(event);
    if(ui->checkBox_password->isChecked())
        ui->edit_password->setEchoMode(QLineEdit::Normal);
    else
        ui->edit_password->setEchoMode(QLineEdit::Password);

    QPainter painter(this);
    painter.setPen(Qt::NoPen);
    if(m_bClicked)
        painter.setBrush(QColor("#d1d1d1"));
    else
        painter.setBrush(QColor("#f0f1f2"));
    painter.drawRoundedRect(rect(), 4, 4);

    updateActiveStatu();
}

bool WifiListWidgetItem::eventFilter(QObject *watched, QEvent *event)
{
    switch (event->type()) {
    case QEvent::KeyPress:
        if(watched == ui->edit_password)
        {
            QKeyEvent* key = static_cast<QKeyEvent*>(event);
            if((key->key() == Qt::Key_Enter || key->key() == Qt::Key_Return) && ui->edit_password->text().size() >= 8)
                onBtnConnectClicked();
        }
        break;
    }
    return QWidget::eventFilter(watched, event);
}

void WifiListWidgetItem::showCurrentLinkStatus(WifiListWidgetItem::ENUM_ShowType type)
{
    switch (type) {
    case ShowInit:
        break;
    case ShowStatus:
        showStatus();
        break;
    case ShowLink:
        showLink();
        break;
    case ShowInputKey:
        showInputKey();
        break;
    case ShowLinking:
        showLinking();
        break;
    case ShowLinkFinished:
        showFinished();
        break;
    }
}

void WifiListWidgetItem::showStatus()
{
    m_showType = ShowStatus;
    checkActive();
    if(m_active == NetworkManager::ActiveConnection::Activating)
    {
//        ui->label_ssid->setText(m_ssid + "   " + tr("Connecting..."));
        ui->label_ssid->setVisible(true);
    }
    else if(m_active == NetworkManager::ActiveConnection::Activated)
    {
        ui->label_ssid->setText(m_ssid + "   " + tr("Connected"));
        ui->label_ssid->setVisible(true);
    }
    else
    {
        ui->label_ssid->setText(m_ssid);
        ui->label_ssid->setVisible(true);
    }
}

void WifiListWidgetItem::showLink()
{
    m_showType = ShowLink;
}

void WifiListWidgetItem::showInputKey()
{
    m_showType = ShowInputKey;
}

void WifiListWidgetItem::showLinking()
{
    m_showType = ShowLinking;
}

void WifiListWidgetItem::showFinished()
{
    m_showType = ShowLinkFinished;
}

void WifiListWidgetItem::connectByPwd(const QString pwd)
{
    NetworkManager::AccessPoint::Ptr access_point = m_network->referenceAccessPoint();
    Q_ASSERT(!access_point.isNull());
    QString dev_uni = m_device->uni();
    QString conn_uni = access_point->uni();
    QString conn_name = access_point->ssid();

    QDBusPendingCallWatcher * watcher = nullptr;
    NetworkManager::Connection::Ptr conn = getConnect();
    if(conn.isNull())
    {
        NMVariantMapMap map_settings;
        QString password = pwd;
        if(password.length() < 8) return;

        NetworkManager::ConnectionSettings::Ptr settings = assembleWpaXPskSettings(access_point, password, true);
        //保存一下密钥
        NMConfig::global()->saveWIFIPasswd(conn_name, password);
        if (settings)
        {
            map_settings = settings->toMap();
        } else {
            //
            return;
        }
        watcher = new QDBusPendingCallWatcher{NetworkManager::addAndActivateConnection(map_settings, dev_uni, conn_uni), this};
    }
    else
    {
        conn_uni = conn->path();
        QString password = pwd;
        if(password.length() < 8) return;
        //更新密码
        NetworkManager::WirelessSecuritySetting::Ptr security_sett
                = conn->settings()->setting(NetworkManager::Setting::WirelessSecurity).dynamicCast<NetworkManager::WirelessSecuritySetting>();
        security_sett->setPsk(password);
        conn->settings()->setAutoconnect(true);
        conn->update(conn->settings()->toMap());
        conn->save();
        //保存一下密钥
        NMConfig::global()->saveWIFIPasswd(conn_name, password);

        watcher = new QDBusPendingCallWatcher{NetworkManager::activateConnection(conn_uni, dev_uni, QString()), this};
    }
    showCurrentLinkStatus(ShowLinking);

    QString dev_name = m_device->interfaceName();
    connect(watcher, &QDBusPendingCallWatcher::finished, [conn_name, dev_name, this] (QDBusPendingCallWatcher * watcher) {
        if (watcher->isError() || !watcher->isValid())
        {
            qWarning() << QStringLiteral("接口%1的连接%2激活失败：%3").arg(dev_name).arg(conn_name)
                          .arg(watcher->error().message());
            //TODO: in what form should we output the warning messages
            showCurrentLinkStatus(ShowLink);
        }
        watcher->deleteLater();
    });
    if(watcher)
        watcher->waitForFinished();
}

void WifiListWidgetItem::connectNoPwd()
{
    NetworkManager::AccessPoint::Ptr access_point = m_network->referenceAccessPoint();
    Q_ASSERT(!access_point.isNull());
    QString dev_uni = m_device->uni();
    QString conn_uni = access_point->uni();
    QString conn_name = access_point->ssid();

    QDBusPendingCallWatcher * watcher;
    NetworkManager::Connection::Ptr conn = getConnect();
    if(conn.isNull())
    {
        NMVariantMapMap map_settings;
        NetworkManager::ConnectionSettings::Ptr settings{new NetworkManager::ConnectionSettings{NetworkManager::ConnectionSettings::Wireless}};
        settings->setId(access_point->ssid());
        settings->setUuid(NetworkManager::ConnectionSettings::createNewUuid());
        settings->setAutoconnect(true);
        map_settings = settings->toMap();
        watcher = new QDBusPendingCallWatcher{NetworkManager::addAndActivateConnection(map_settings, dev_uni, conn_uni), this};
    }
    else
    {
        conn_uni = conn->path();
        conn->settings()->setAutoconnect(true);
        conn->update(conn->settings()->toMap());
        conn->save();
        watcher = new QDBusPendingCallWatcher{NetworkManager::activateConnection(conn_uni, dev_uni, QString()), this};
    }

    //
    showCurrentLinkStatus(ShowLinking);

    QString dev_name = m_device->interfaceName();
    connect(watcher, &QDBusPendingCallWatcher::finished, [conn_name, dev_name, this] (QDBusPendingCallWatcher * watcher) {
        if (watcher->isError() || !watcher->isValid())
        {
            qWarning() << QStringLiteral("接口%1的连接%2激活失败：%3").arg(dev_name).arg(conn_name)
                          .arg(watcher->error().message());
            //TODO: in what form should we output the warning messages
            showCurrentLinkStatus(ShowLink);
//            QMessageBox::warning(this, "waring", "Connect Fail");
        }
        watcher->deleteLater();
    });
}

QString WifiListWidgetItem::getIconBySignalStrength(int strength)
{
    if (0 >= strength)
        return QStringLiteral(":/image/no-signal.png");
    else if (25 >= strength)
        return QStringLiteral(":/image/signal-1.png");
    else if (50 >= strength)
        return QStringLiteral(":/image/signal-2.png");
    else if (75 >= strength)
        return QStringLiteral(":/image/signal-3.png");
    else
        return QStringLiteral(":/image/signal-4.png");
}

NetworkManager::ActiveConnection::State WifiListWidgetItem::checkActive()
{
    m_active = NetworkManager::ActiveConnection::Unknown;
    for(NetworkManager::ActiveConnection::Ptr aConn: NetworkManager::activeConnections())
    {
        if(aConn->id() == m_network->ssid())
        {
            m_active = aConn->state();
            if(m_active == NetworkManager::ActiveConnection::Activated)
            {
                ui->label_pwd->setVisible(false);
            }
            else {
                ui->label_pwd->setVisible(true);
            }
            break;
        }
    }
    if(m_active ==  NetworkManager::ActiveConnection::Activating) {
        if(!mConnectingTimer->isActive()) mConnectingTimer->start(); }
    else {
        mConnectingTimer->stop(); }

    return m_active;
}

bool WifiListWidgetItem::checkKnown()
{
    m_known = false;
    NetworkManager::Connection::Ptr conn = getConnect();
    if(conn && conn->isValid())
    {
        m_known = true;
    }
    return m_known;
}

bool WifiListWidgetItem::checkAuto()
{
    m_auto = true;
    NetworkManager::Connection::Ptr conn = getConnect();
    if(conn && conn->isValid())
    {
        m_auto =  conn->settings()->autoconnect();
    }
    return m_auto;
}

bool WifiListWidgetItem::checkNeedPassword(QString &password)
{
    m_needPassword = false;
    NetworkManager::AccessPoint::Ptr access_point = m_network->referenceAccessPoint();
    Q_ASSERT(!access_point.isNull());
    NetworkManager::WirelessSecurityType sec_type = NetworkManager::findBestWirelessSecurity(m_device->wirelessCapabilities()
            , true, (m_device->mode() == NetworkManager::WirelessDevice::Adhoc)
            , access_point->capabilities(), access_point->wpaFlags(), access_point->rsnFlags());
    m_secType = sec_type;
    switch (sec_type)
    {
    case NetworkManager::WpaPsk:
    case NetworkManager::Wpa2Psk:
    {
        m_needPassword = true;
        break;
    }
    default:
        break;
    }

    NetworkManager::Connection::Ptr conn = getConnect();
    if(conn && conn->isValid())
    {
        password = NMConfig::global()->readWIFIPasswd(conn->name());
    }
    return m_needPassword;
}

NetworkManager::Connection::Ptr WifiListWidgetItem::getConnect()
{
    QString conn_name = m_network->ssid();
    NetworkManager::Connection::Ptr conn;
    for (auto const & dev_conn : m_device->availableConnections())
    {
        if (dev_conn->settings()->id() == conn_name)
        {
            conn = dev_conn;
        }
    }
    return conn;
}
